import functools

from django.shortcuts import redirect

from .handlers import AuthHandler


def azure_auth_required(func):
    @functools.wraps(func)
    def _wrapper(request, *args, **kwargs):

        # If the token is valid (or a new valid one can be generated)
        if AuthHandler(request).get_token_from_cache():
            # If the user is authenticated
            if request.user.is_authenticated:
                return func(request, *args, **kwargs)
        return redirect("azure_auth:login")

    return _wrapper



import pytest
from django.contrib.auth import get_user_model
from mixer.backend.django import mixer

UserModel = get_user_model()


@pytest.fixture(scope="function", autouse=True)
def user(request):
    _user = mixer.blend(UserModel)
    _user.username = _user.email
    _user.save()

    if request.cls:
        request.cls.user = _user
    return _user


@pytest.fixture(scope="function")
def auth_flow(request):
    _auth_flow = {
        "state": "dummy_state",
        "redirect_uri": "http://localhost:8000/azure_auth/callback",
        "scope": ["profile", "offline_access", "User.Read", "openid"],
        "auth_uri": "https://aad-auth-uri.com",
    }

    if request.cls:
        request.cls.auth_flow = _auth_flow
    return _auth_flow


@pytest.fixture(scope="function")
def token(request):
    _token = {
        "token_type": "Bearer",
        "scope": "openid profile User.Read email",
        "expires_in": 3749,
        "ext_expires_in": 3749,
        "access_token": "dummy_access_token",
        "refresh_token": "dummy_refresh_token",
        "id_token": "dummy_id_token",
        "client_info": "dummy_client_info",
    }

    if request.cls:
        request.cls.token = _token
    return _token



from http import HTTPStatus

import msal
import requests
from django.conf import settings
from django.contrib.auth import get_user_model

from azure_auth.exceptions import DjangoAzureAuthException, TokenError

UserModel = get_user_model()


class AuthHandler:
    """
    Class to interface with `msal` package and execute authentication process.
    """

    def __init__(self, request=None):
        """

        :param request: HttpRequest
        """
        self.request = request
        self.graph_user_endpoint = "https://graph.microsoft.com/v1.0/me"
        self.auth_flow_session_key = "auth_flow"
        self._cache = msal.SerializableTokenCache()
        self._msal_app = None

    def get_auth_uri(self) -> str:
        """
        Requests the auth flow dictionary and stores it on the session to be
        queried later in the auth process.

        :return: Authentication redirect URI
        """
        # TODO: Handle if user has put incorrect details in settings
        flow = self.msal_app.initiate_auth_code_flow(
            scopes=settings.AZURE_AUTH["SCOPES"],
            redirect_uri=settings.AZURE_AUTH["REDIRECT_URI"],
        )
        self.request.session[self.auth_flow_session_key] = flow
        return flow["auth_uri"]

    def get_token_from_flow(self) -> dict:
        """
        Acquires the token from the auth flow on the session and the content of
        the redirect request from Active Directory.

        :return: Token result containing `access_token`/`id_token` and other
        claims, depending on scopes used
        """
        flow = self.request.session.pop(self.auth_flow_session_key, {})
        token_result = self.msal_app.acquire_token_by_auth_code_flow(
            auth_code_flow=flow, auth_response=self.request.GET
        )
        if "error" in token_result:
            raise TokenError(token_result["error"], token_result["error_description"])
        self._save_cache()
        return token_result

    def get_token_from_cache(self):
        accounts = self.msal_app.get_accounts()
        if accounts:  # pragma: no branch
            # Will return `None` if CCA cannot retrieve or generate new token
            token_result = self.msal_app.acquire_token_silent(
                scopes=settings.AZURE_AUTH["SCOPES"], account=accounts[0]
            )
            self._save_cache()
            return token_result

    def authenticate(self, token: dict) -> UserModel:
        """
        Helper method to authenticate the user. Gets the Azure user from the
        Microsoft Graph endpoint and gets/creates the associated Django user.

        :param token: MSAL auth token dictionary
        :return: Django user instance
        """
        azure_user = self._get_azure_user(token["access_token"])

        # Allow for `outlook.com` users with email set on the
        # `userPrincipalName` attribute
        email = (
            azure_user["mail"]
            if azure_user.get("mail", None)
            else azure_user["userPrincipalName"]
        )

        # Using `UserModel._default_manager.get_by_natural_key` handles custom
        # user model and `USERNAME_FIELD` setting
        try:
            user = UserModel._default_manager.get_by_natural_key(email)
        except UserModel.DoesNotExist:
            user = UserModel._default_manager.create_user(username=email, email=email)
            user.first_name = attr if (attr := azure_user["givenName"]) else ""
            user.last_name = attr if (attr := azure_user["surname"]) else ""
            user.save()

        # TODO: Handle groups
        return user

    @staticmethod
    def get_logout_uri() -> str:
        """
        Forms the URI to log the user out in the Active Directory app and
        redirect to the webapp logout page.

        :return: Active Directory app logout URI
        """
        authority = settings.AZURE_AUTH["AUTHORITY"]
        logout_uri = settings.AZURE_AUTH.get("LOGOUT_URI", "")
        if logout_uri:
            return (
                f"{authority}/oauth2/v2.0/logout?post_logout_redirect_uri={logout_uri}"
            )
        return f"{authority}/oauth2/v2.0/logout"

    @property
    def msal_app(self):
        if self._msal_app is None:
            self._msal_app = msal.ConfidentialClientApplication(
                client_id=settings.AZURE_AUTH["CLIENT_ID"],
                client_credential=settings.AZURE_AUTH["CLIENT_SECRET"],
                authority=settings.AZURE_AUTH["AUTHORITY"],
                token_cache=self.cache,
            )
        return self._msal_app

    @property
    def cache(self):
        if self.request.session.get("token_cache"):
            self._cache.deserialize(self.request.session["token_cache"])
        return self._cache

    def _save_cache(self):
        if self.cache.has_state_changed:
            self.request.session["token_cache"] = self.cache.serialize()

    def _get_azure_user(self, token: str):
        resp = requests.get(
            self.graph_user_endpoint, headers={"Authorization": f"Bearer {token}"}
        )
        if resp.ok:
            return resp.json()
        elif resp.status_code == HTTPStatus.UNAUTHORIZED:
            error = resp.json()["error"]
            raise TokenError(message=error["code"], description=error["message"])
        else:  # pragma: no cover
            raise DjangoAzureAuthException("An unknown error occurred.")



from django.conf import settings
from django.contrib.auth import authenticate, login, logout
from django.http import HttpResponseForbidden, HttpResponseRedirect

from .handlers import AuthHandler


def azure_auth_login(request):
    return HttpResponseRedirect(AuthHandler(request).get_auth_uri())


def azure_auth_logout(request):
    logout(request)
    return HttpResponseRedirect(AuthHandler.get_logout_uri())


def azure_auth_callback(request):
    token = AuthHandler(request).get_token_from_flow()
    user = authenticate(request, token=token)
    if user:
        login(request, user)
    else:
        return HttpResponseForbidden("Invalid email for this app.")
    return HttpResponseRedirect(settings.LOGIN_REDIRECT_URL)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['azure_auth', 'azure_auth.tests']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2', 'msal>=1.18.0,<2.0.0']

setup_kwargs = {
    'name': 'django-azure-auth',
    'version': '1.0.1',
    'description': 'A simple Django app for user authentication with Azure Active Directory.',
    'long_description': '![Build](https://github.com/AgileTek/django-azure-auth/actions/workflows/push-actions.yml/badge.svg)\n\n# Django Azure Auth\nA simple Django app for user authentication with Azure Active Directory.\n\n## Description\n`django-azure-auth` is a Django app which wraps the great [MSAL](https://github.com/AzureAD/microsoft-authentication-library-for-python)\npackage to enable authentication against Microsoft\'s Azure Active Directory in Django projects.\n\nThe app includes `login`, `logout` and `callback` authentication views, a decorator\nto protect individual views, and middleware which allows the entire site to require user \nauthentication by default, with the ability to exempt specified views.\n\nThis project is in no way affiliated with Microsoft.\n\n## Installation\nFrom PyPi:\n```bash\npip install django-azure-auth\n```\n\n## Configuration\n### Azure setup\n- Register an app at https://portal.azure.com/.\n- Add a client secret and note it down.\n- Add a redirect URI of the format `https://<domain>/azure_auth/callback`.\n\n### Settings\nAdd the following to your `settings.py`, replacing the variables in braces with the values\nfrom your Azure app: \n```python\nAZURE_AUTH = {\n    "CLIENT_ID": "<client id>",\n    "CLIENT_SECRET": "<client secret>",\n    "REDIRECT_URI": "https://<domain>/azure_auth/callback",\n    "SCOPES": ["User.Read"],\n    "AUTHORITY": "https://login.microsoftonline.com/<tenant id>",   # Or https://login.microsoftonline.com/common if multi-tenant\n    "LOGOUT_URI": "https://<domain>/logout",    # Optional\n    "PUBLIC_URLS": ["<public:view_name>",]  # Optional, public views accessible by non-authenticated users\n}\nLOGIN_URL = "/azure_auth/login"\nLOGIN_REDIRECT_URL = "/"    # Or any other endpoint\n```\n#### Note: You should obfuscate the credentials by using environment variables.\n\n### Installed apps\nAdd the following to your `INSTALLED_APPS`:\n```python\nINSTALLED_APPS = (\n    "...",\n    "azure_auth",\n    "..."\n)\n```\n\n### Authentication backend\nConfigure the authentication backend:\n```python\nAUTHENTICATION_BACKENDS = ("azure_auth.backends.AzureBackend",)\n```\n\n### URLs\nInclude the app\'s URLs in your `urlpatterns`:\n```python\nfrom django.urls import path, include\n\nurlpatterns = [\n    path("azure_auth/", include("azure_auth.urls"),),\n]\n```\n\n## Usage\n### Decorator\nTo make user authentication a requirement for accessing an individual view, decorate the\nview like so:\n```python\nfrom azure_auth.decorators import azure_auth_required\nfrom django.shortcuts import HttpResponse\n\n@azure_auth_required\ndef protected_view(request):\n    return HttpResponse("A view protected by the decorator")\n```\n\n### Middleware\nIf you want to protect your entire site by default, you can use the middleware by adding the \nfollowing to your `settings.py`:\n```python\nMIDDLEWARE = [\n    "...",\n    "azure_auth.middleware.AzureMiddleware",\n    "...",\n]\n```\nMake sure you add the middleware after Django\'s `session` and `authentication` middlewares so \nthat the request includes the session and user objects. Public URLs which need to be accessed by \nnon-authenticated users should be specified in the `settings.AZURE_AUTH["PUBLIC_URLS"]`, as \nshown above.\n\n## Planned development\n- Groups management\n\n## Credits\nThis app is heavily inspired by and builds on functionality in \nhttps://github.com/shubhamdipt/django-microsoft-authentication, with both feature \nimprovements and code assurance through testing.\n\nCredit also to:\n- https://github.com/Azure-Samples/ms-identity-python-webapp\n- https://github.com/AzMoo/django-okta-auth',
    'author': 'AgileTek Engineering',
    'author_email': 'london@agiletek.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AgileTek/django-azure-auth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from django.conf import settings
from django.shortcuts import redirect
from django.urls import reverse

from .handlers import AuthHandler


class AzureMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        public_views = ["azure_auth:login", "azure_auth:logout", "azure_auth:callback"]
        public_views.extend(settings.AZURE_AUTH.get("PUBLIC_URLS", []))
        public_urls = [reverse(view_name) for view_name in public_views]

        if request.path_info in public_urls:
            return self.get_response(request)

        if AuthHandler(request).get_token_from_cache():
            # If the user is authenticated
            if request.user.is_authenticated:
                return self.get_response(request)
        return redirect("azure_auth:login")
